
"""
This is the facade of all features of DDF.
Module that contains wrappers and shortcuts (aliases).
"""
import warnings

from django.apps import apps

from django_dynamic_fixture.ddf import DynamicFixture, Copier, Mask, DDFLibrary, \
    set_pre_save_receiver, set_post_save_receiver
from django_dynamic_fixture.django_helper import print_field_values, django_greater_than
from django_dynamic_fixture.fixture_algorithms.sequential_fixture import SequentialDataFixture, \
    StaticSequentialDataFixture
from django_dynamic_fixture.global_settings import DDF_DEFAULT_DATA_FIXTURE, DDF_FILL_NULLABLE_FIELDS, DDF_FK_MIN_DEPTH, \
                                                    DDF_IGNORE_FIELDS, DDF_VALIDATE_MODELS, \
                                                    DDF_DEBUG_MODE, DDF_FIELD_FIXTURES
from django_dynamic_fixture.script_ddf_checkings import ddf_check_models


__version__ = '4.0.1'


if not django_greater_than(4, 0):
    warnings.warn("DDF 4.* officially supports only Django 4 or higher.", DeprecationWarning)


LOOKUP_SEP = '__'

def look_up_alias(ddf_as_f=True, **kwargs):
    """
    Example of parameters:
    a__b__c=1 => {a: {b: {c: 1}}}
    """
    field_dict = {}
    params_to_be_skipped = []
    for alias, value in kwargs.items():
        parts = alias.split(LOOKUP_SEP)
        if len(parts) == 1:
            params_to_be_skipped.append(alias)
        level_dict = field_dict
        for part in parts[:-1]:
            level_dict = level_dict.setdefault(part, {})
        level_dict[parts[-1]] = value
    if ddf_as_f:
        for root, value in field_dict.items():
            if root in params_to_be_skipped:
                field_dict[root] = value
            else:
                field_dict[root] = dict_to_f(value)
    return field_dict


def dict_to_f(value):
    """
    Example:
    1 => 1
    {b: 1} => F(b=1)
    {b: {c: 1}} => F(b=F(c=1))
    """
    if not isinstance(value, dict):
        return value
    else:
        kwargs = {}
        for k, v in value.items():
            if not isinstance(v, dict):
                kwargs[k] = v
            else:
                kwargs[k] = dict_to_f(v)
        return F(**kwargs)


def fixture(**kwargs):
    """
    DynamicFixture factory: It instantiate a DynamicFixture using global configurations.
    Same as F(...)
    """
    kwargs = look_up_alias(**kwargs)
    f = DynamicFixture(data_fixture=kwargs.pop('data_fixture', DDF_DEFAULT_DATA_FIXTURE),
                       fill_nullable_fields=kwargs.pop('fill_nullable_fields', DDF_FILL_NULLABLE_FIELDS),
                       ignore_fields=kwargs.pop('ignore_fields', []),
                       fk_min_depth=kwargs.pop('fk_min_depth', DDF_FK_MIN_DEPTH),
                       validate_models=kwargs.pop('validate_models', DDF_VALIDATE_MODELS),
                       print_errors=kwargs.pop('print_errors', True),
                       debug_mode=kwargs.pop('debug_mode', DDF_DEBUG_MODE),
                       **kwargs)
    return f


# Wrappers

def _new(model, n=1, ddf_lesson=None, persist_dependencies=False, **kwargs):
    """
    Return one or many valid instances of Django Models with fields filled with auto generated or customized data.
    All instances will NOT be persisted in the database, except its dependencies, in case @persist_dependencies is True.

    @model: The class of the Django model. It can be a string `<app_label>.<model_name>`
    @n: number of instances to be created with the given configuration. Default is 1.
    @ddf_lesson: use a custom ddf_lesson to build the model object.
    @persist_dependencies: If True, save internal dependencies, otherwise just instantiate them. Default is False.

    @data_fixture: override DDF_DEFAULT_DATA_FIXTURE configuration. Default is SequentialDataFixture().
    @fill_nullable_fields: override DDF_FILL_NULLABLE_FIELDS global configuration. Default is True.
    @ignore_fields: List of fields that will be ignored by DDF. It will be concatenated with the global list DDF_IGNORE_FIELDS. Default is [].
    @fk_min_depth: override DDF_FK_MIN_DEPTH global configuration. Default 0.
    @validate_models: override DDF_VALIDATE_MODELS global configuration. Default is False.
    @print_errors: print on console all instance values if DDF can not generate a valid object with the given configuration.

    Wrapper for the method DynamicFixture.new
    """
    if isinstance(model, str):
        model = apps.get_model(model)
    kwargs = look_up_alias(**kwargs)
    d = fixture(**kwargs)
    if n == 1:
        return d.new(model, ddf_lesson=ddf_lesson, persist_dependencies=persist_dependencies, **kwargs)
    instances = []
    for _ in range(n):
        instances.append(d.new(model, ddf_lesson=ddf_lesson, persist_dependencies=persist_dependencies, **kwargs))
    return instances


def _get(model, n=1, ddf_lesson=None, **kwargs):
    """
    Return one or many valid instances of Django Models with fields filled with auto generated or customized data.
    All instances will be persisted in the database.

    @model: The class of the Django model. It can be a string `<app_label>.<model_name>`
    @n: number of instances to be created with the given configuration. Default is 1.
    @ddf_lesson: use a custom ddf_lesson to build the model object.

    @data_fixture: override DDF_DEFAULT_DATA_FIXTURE configuration. Default is SequentialDataFixture().
    @fill_nullable_fields: override DDF_FILL_NULLABLE_FIELDS global configuration. Default is True.
    @ignore_fields: List of fields that will be ignored by DDF. It will be concatenated with the global list DDF_IGNORE_FIELDS. Default is [].
    @fk_min_depth: override DDF_FK_MIN_DEPTH global configuration. Default 0.
    @validate_models: override DDF_VALIDATE_MODELS global configuration. Default is False.
    @print_errors: print on console all instance values if DDF can not generate a valid object with the given configuration.

    Wrapper for the method DynamicFixture.get
    """
    if isinstance(model, str):
        model = apps.get_model(model)
    kwargs = look_up_alias(**kwargs)
    d = fixture(**kwargs)
    if n == 1:
        return d.get(model, ddf_lesson=ddf_lesson, **kwargs)
    instances = []
    for _ in range(n):
        instances.append(d.get(model, ddf_lesson=ddf_lesson, **kwargs))
    return instances


def _teach(model, ddf_lesson=None, **kwargs):
    '''
    @model: The class of the Django model. It can be a string `<app_label>.<model_name>`
    @ddf_lesson: Name of custom ddf_lesson to be created.

    @raise an CantOverrideddf_lesson error if the same model/ddf_lesson were called twice.

    Sometimes DDF can't create an model instance because the particularities of the model.
    The workaround for this is to teach DDF how to create it.
    Basically, we use the same object creation approach that will be saved as a template
    for the next DDF calls.

    Use this method to teach DDF how to create an instance.

    New metaphor for the shelve/library feature.
    `Shelve` becomes `Teach`
    `Library` becomes `Lessons`
    '''
    if isinstance(model, str):
        model = apps.get_model(model)
    kwargs = look_up_alias(**kwargs)
    d = fixture(**kwargs)
    return d.teach(model, ddf_lesson=ddf_lesson, **kwargs)


# Shortcuts
N = new = _new
G = get = _get
T = teach = _teach
F = fixture
C = Copier
M = Mask
P = print_field_values
DDFLibrary = DDFLibrary
PRE_SAVE = set_pre_save_receiver
POST_SAVE = set_post_save_receiver

import typing

INSTANCE_TYPE = typing.TypeVar('INSTANCE')

def new(model: typing.Type[INSTANCE_TYPE], n: int = 1, ddf_lesson = None, persist_dependencies: bool = True, **kwargs) -> INSTANCE_TYPE:
    return _new(model, n=n, ddf_lesson=ddf_lesson, persist_dependencies=persist_dependencies, **kwargs)

def get(model: typing.Type[INSTANCE_TYPE], n: int=1, ddf_lesson=None, **kwargs) -> INSTANCE_TYPE:
    return _get(model, n=n, ddf_lesson=ddf_lesson, **kwargs)

def teach(model: typing.Type[INSTANCE_TYPE], ddf_lesson=None, **kwargs):
    return _teach(model, ddf_lesson=ddf_lesson, **kwargs)

N = new
G = get
T = teach



from django.db import models


class Publisher(models.Model):
    name = models.CharField(max_length=100, unique=True)

class Author(models.Model):
    name = models.CharField(max_length=100)
    description = models.TextField(null=True)

class Category(models.Model):
    name = models.CharField(max_length=100, unique=True)
    parent = models.ForeignKey('self', on_delete=models.DO_NOTHING, null=True, blank=True)

class Book(models.Model):
    isb = models.CharField(max_length=100, unique=True)
    name = models.CharField(max_length=100)
    main_author = models.ForeignKey(Author, related_name='books', on_delete=models.DO_NOTHING)
    authors = models.ManyToManyField('Author', related_name='m2m')
    categories = models.ManyToManyField('Category', related_name='m2m')
    from .fields import JSONField
    metadata = JSONField(null=True)

class BookPublisher(models.Model):
    book_edition = models.ForeignKey('BookEdition', on_delete=models.DO_NOTHING)
    publisher = models.ForeignKey('Publisher', on_delete=models.DO_NOTHING)
    comments = models.TextField(max_length=100)

class BookEdition(models.Model):
    book = models.ForeignKey(Book, related_name='editions', on_delete=models.DO_NOTHING)
    publishers = models.ManyToManyField('Publisher', related_name='edition_publishers', through=BookPublisher)
    year = models.IntegerField()



import pytest

from django.test import TestCase

from ddf import *


class SampleAppTestCase(TestCase):
    def test_publisher(self):
        o = G('django_dynamic_fixture.Publisher')
        assert o.name is not None

    def test_author(self):
        o = G('django_dynamic_fixture.Author')
        assert o.name is not None
        assert o.description is None
        o = G('django_dynamic_fixture.Author', fill_nullable_fields=True)
        assert o.description is not None

    def test_category(self):
        o = G('django_dynamic_fixture.Category')
        assert o.name is not None
        assert o.parent is None
        o = G('django_dynamic_fixture.Category', fk_min_depth=1)
        assert o.parent is not None
        assert o.parent.parent is None
        o = G('django_dynamic_fixture.Category', fk_min_depth=2)
        assert o.parent.parent is not None
        assert o.parent.parent.parent is None

    def test_book(self):
        o = G('django_dynamic_fixture.Book')
        assert o.main_author is not None
        assert o.main_author.name is not None
        assert o.authors.all().count() == 0
        assert o.categories.all().count() == 0
        o = G('django_dynamic_fixture.Book', authors=2, categories=5)
        assert o.authors.all().count() == 2
        assert o.categories.all().count() == 5
        o = G('django_dynamic_fixture.Book', metadata={'a': 1})
        assert o.metadata == {'a': 1}

    def test_book_edition(self):
        o = G('django_dynamic_fixture.BookEdition')
        assert o.book is not None
        assert o.year is not None
        assert o.publishers.all().count() == 0
        o = G('django_dynamic_fixture.BookEdition', publishers=2)
        assert o.publishers.all().count() == 2

    def test_book_publisher(self):
        o = G('django_dynamic_fixture.BookPublisher')
        assert o.book_edition.book is not None
        assert o.publisher is not None
        assert o.comments is not None



import uuid

from django.db import models
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


try:
    from django.contrib.gis.geos import *
    from django.contrib.gis.db import models as geomodels
except ImproperlyConfigured:
    pass  # environment without geo libs


from django.test import TestCase

from django_dynamic_fixture.fixture_algorithms.default_fixture import BaseDataFixture


class BaseDataFixtureTestCase(TestCase):
    def setUp(self):
        self.fixture = BaseDataFixture()

    def test_uuid(self):
        assert isinstance(self.fixture.generate_data(models.UUIDField()), uuid.UUID)


if (hasattr(settings, 'DDF_TEST_GEODJANGO') and settings.DDF_TEST_GEODJANGO):
    from django_dynamic_fixture.fixture_algorithms.default_fixture import GeoDjangoFixtureMixin

    # Mixing for tests
    class GeoDjangoFixtureMixin(BaseDataFixture, GeoDjangoFixtureMixin):
        pass

    class GeoDjangoDataFixtureTestCase(TestCase):
        def setUp(self):
            self.fixture = GeoDjangoFixtureMixin()

        def test_geometryfield_config(self):
            assert isinstance(self.fixture.generate_data(geomodels.GeometryField()), GEOSGeometry)

        def test_pointfield_config(self):
            assert isinstance(self.fixture.generate_data(geomodels.PointField()), Point)

        def test_linestringfield_config(self):
            assert isinstance(self.fixture.generate_data(geomodels.LineStringField()), LineString)

        def test_polygonfield_config(self):
            assert isinstance(self.fixture.generate_data(geomodels.PolygonField()), Polygon)

        def test_multipointfield_config(self):
            assert isinstance(self.fixture.generate_data(geomodels.MultiPointField()), MultiPoint)

        def test_multilinesstringfield_config(self):
            assert isinstance(self.fixture.generate_data(geomodels.MultiLineStringField()), MultiLineString)

        def test_multipolygonfield_config(self):
            assert isinstance(self.fixture.generate_data(geomodels.MultiPolygonField()), MultiPolygon)

        def test_geometrycollectionfield_config(self):
            assert isinstance(self.fixture.generate_data(geomodels.GeometryCollectionField()), GeometryCollection)



from django.db import models


try:
    from django.contrib.postgres.fields import JSONField
    class ModelForPostgresFields(models.Model):
        nullable_json_field = JSONField(null=True)
        json_field = JSONField(null=False)
        class Meta:
            app_label = 'django_dynamic_fixture'
except ImportError:
    pass

try:
    from jsonfield import JSONField
    from jsonfield import JSONCharField
    class ModelForPlugins1(models.Model):
        json_field1 = JSONCharField(max_length=10)
        json_field2 = JSONField()
        class Meta:
            app_label = 'django_dynamic_fixture'
except ImportError:
    print('Library `jsonfield` not installed. Skipping.')


try:
    from json_field import JSONField as JSONField2
    class ModelForPlugins2(models.Model):
        json_field1 = JSONField2()
        class Meta:
            app_label = 'django_dynamic_fixture'
except ImportError:
    print('Library `django-json-field` not installed. Skipping.')


try:
    from polymorphic.models import PolymorphicModel
    class ModelPolymorphic(PolymorphicModel):
        class Meta:
            verbose_name = 'Polymorphic Model'


    class ModelPolymorphic2(ModelPolymorphic):
        class Meta:
            verbose_name = 'Polymorphic Model 2'


    class ModelPolymorphic3(ModelPolymorphic):
        class CannotSave(Exception):
            pass

        def save(self):
            raise self.CannotSave
except ImportError:
    print('Library `django_polymorphic` not installed. Skipping.')




import os
import tempfile

from shutil import rmtree, copy2
from django.core.files import File

from django.test import TestCase
from django.conf import settings
from django.core.files.storage import FileSystemStorage


TEMP_PATH = tempfile.gettempdir() or os.environ.get('TEMP')
TEMP_PATH_DDF = os.path.join(TEMP_PATH, 'DDF_TEMP')


class CustomFileSystemStorage(FileSystemStorage):
    def __init__(self, *args, **kwargs):
        super().\
        __init__(location=TEMP_PATH_DDF, *args, **kwargs)


class FileSystemDjangoTestCase(TestCase):
    TEAR_DOWN_ENABLED = True

    def setUp(self):
        self.fdf_setup()

    def tearDown(self):
        self.fdf_teardown()

    def _pre_setup(self):
        super()._pre_setup()
        self.fdf_setup()

    def _post_teardown(self):
        "Try to remove all files and directories created by the test."
        super()._post_teardown()
        self.fdf_teardown()

    def fdf_setup(self):
        self.directories = []
        self.files = {}
        setattr(settings, 'DEFAULT_FILE_STORAGE', 'django_dynamic_fixture.fdf.CustomFileSystemStorage')

    def fdf_teardown(self):
        if self.TEAR_DOWN_ENABLED:
            while self.files:
                self.remove_temp_file(next(iter(self.files.keys())))
            while self.directories:
                self.remove_temp_directory(self.directories[0])
            if os.path.exists(TEMP_PATH_DDF):
                rmtree(TEMP_PATH_DDF)

    def create_temp_directory(self, prefix='file_system_test_case_dir_'):
        "Create a temporary directory and returns the directory pathname."
        directory = tempfile.mkdtemp(prefix=prefix)
        self.directories.append(directory)
        return directory

    def remove_temp_directory(self, directory_pathname):
        "Remove a directory."
        rmtree(directory_pathname)
        if directory_pathname in self.directories:
            try:
                self.directories.remove(directory_pathname)
            except WindowsError:
                pass

    def create_temp_file(self, directory=None, prefix='file_system_test_case_file_', suffix='.tmp'):
        """
        Create a temporary file with a option prefix and suffix in a temporary or custom directory.
        Returns the filepath
        """
        tmp_file = tempfile.mkstemp(prefix=prefix, dir=directory, suffix=suffix)
        file_obj = os.fdopen(tmp_file[0])
        self.files[tmp_file[1]] = file_obj
        return tmp_file[1]

    def create_temp_file_with_name(self, directory, name):
        "Create a temporary file with a specified name."
        filepath = os.path.join(directory, name)
        file_obj = open(filepath, 'wb')
        file_obj.close()
        self.files[filepath] = file_obj
        return filepath

    def rename_temp_file(self, filepath, name):
        "Rename an existent file. 'name' is not a file path, so it must not include the directory path name."
        directory = self.get_directory_of_the_file(filepath)
        new_filepath = os.path.join(directory, name)
        os.rename(filepath, new_filepath)
        if filepath in self.files.keys():
            self.files.pop(filepath)
        self.files[new_filepath] = open(new_filepath, 'a+b')
        self.files[new_filepath].close()
        return new_filepath

    def remove_temp_file(self, filepath):
        "Remove a file."
        if filepath in self.files.keys():
            fileobj = self.files.pop(filepath)
            fileobj.close()
        if os.path.exists(filepath):
            try:
                os.unlink(filepath)
            except WindowsError:
                pass

    def copy_file_to_dir(self, filepath, directory):
        "Copy a file to a specified directory."
        copy2(filepath, directory)
        return self.get_filepath(directory, self.get_filename(filepath))

    def add_text_to_file(self, filepath, content):
        "Add text to an existent file."
        file = open(filepath, 'a')
        file.write(content)
        file.close()

    def get_directory_of_the_file(self, filepath):
        "Get the directory path name of a file."
        return os.path.dirname(filepath)

    def get_filename(self, filepath):
        "Get the filename of a file."
        return os.path.basename(filepath)

    def get_filepath(self, directory, filename):
        "Get the file path of a file with a defined name in a directory."
        return os.path.join(directory, filename)

    def get_content_of_file(self, filepath):
        "Returns the content of a file."
        file = open(filepath, 'r')
        content = file.read()
        file.close()
        return content

    def create_django_file_with_temp_file(self, name, content=None, dir=None, prefix='file_system_test_case_file_', suffix='.tmp'):
        "Create and returns a django.core.files.File"
        file = open(self.create_temp_file(directory=dir, prefix=prefix, suffix=suffix), 'w')
        file.close()
        django_file = File(file, name=name)
        self.files[django_file.file.name] = open(django_file.file.name, 'a+b')
        if content:
            self.files[django_file.file.name].write(content)
        self.files[django_file.file.name].close()
        return django_file

    def create_django_file_using_file(self, filepath):
        "Create and returns a django.core.files.File"
        new_filepath = self.copy_file_to_dir(filepath, self.create_temp_directory())
        the_file = open(new_filepath, 'rb')
        django_file = File(the_file, name=os.path.basename(new_filepath))
        self.files[django_file.file.name] = the_file
        #self.files[django_file.file.name].close()
        return django_file

    def assertFileExists(self, filepath):
        assert os.path.exists(filepath), '%s does not exist' % filepath

    def assertFileDoesNotExists(self, filepath):
        assert os.path.exists(filepath) is False, '%s exist' % filepath

    def assertDirectoryExists(self, directory):
        "@directory must be the directory path"
        assert os.path.exists(directory), '%s does not exist' % directory

    def assertDirectoryDoesNotExists(self, directory):
        "@directory must be the directory path"
        assert os.path.exists(directory) is False, '%s exist' % directory

    def assertDirectoryContainsFile(self, directory, filename):
        filepath = os.path.join(directory, filename)
        self.assertFileExists(filepath)

    def assertDirectoryDoesNotContainsFile(self, directory, filename):
        filepath = os.path.join(directory, filename)
        self.assertFileDoesNotExists(filepath)

    def assertFilesHaveEqualLastModificationTimestamps(self, filepath1, filepath2):
        assert os.path.getmtime(filepath1) - os.path.getmtime(filepath2) == 0

    def assertFilesHaveNotEqualLastModificationTimestamps(self, filepath1, filepath2):
        assert os.path.getmtime(filepath1) - os.path.getmtime(filepath2) != 0

    def assertNumberOfFiles(self, directory, number_of_files):
        filenames = [filename for filename in os.listdir(directory) if os.path.isfile(os.path.join(directory, filename))]
        assert len(filenames) == number_of_files, '[%s] %s' % (len(filenames), filenames)




from django.test import TestCase
from django.db import models
import pytest

from django_dynamic_fixture import N, G
from django_dynamic_fixture.models_test import *
from django_dynamic_fixture.django_helper import *


class DjangoHelperAppsTest(TestCase):
    def test_get_apps_must_return_all_installed_apps(self):
        assert len(get_apps()) >= 1

    def test_get_apps_may_be_filtered_by_app_names(self):
        apps = get_apps(application_labels=['django_dynamic_fixture'])
        assert len(apps) == 1

    def test_get_apps_may_ignore_some_apps(self):
        apps = len(get_apps(exclude_application_labels=['django_dynamic_fixture']))
        assert len(get_apps()) - apps == 1

    def test_app_name_must_be_valid(self):
        with pytest.raises(Exception):
            get_apps(application_labels=['x'])
        with pytest.raises(Exception):
            get_apps(exclude_application_labels=['x'])

    def test_get_app_name_must(self):
        import django_dynamic_fixture.models as ddf
        assert get_app_name(ddf) == 'django_dynamic_fixture'

    def test_get_models_of_an_app_must(self):
        ddf = get_apps(application_labels=['django_dynamic_fixture'])[0]
        models_ddf = get_models_of_an_app(ddf)
        assert len(models_ddf) > 0
        assert ModelWithNumbers in models_ddf


class DjangoHelperModelsTest(TestCase):
    def test_get_model_name(self):
        class MyModel_test_get_model_name(models.Model): pass
        assert get_model_name(MyModel_test_get_model_name) == 'MyModel_test_get_model_name'

    def test_get_unique_model_name(self):
        class MyModel_test_get_unique_model_name(models.Model): pass
        assert get_unique_model_name(MyModel_test_get_unique_model_name) == 'django_dynamic_fixture.tests.test_django_helper.MyModel_test_get_unique_model_name'

    def test_get_fields_from_model(self):
        class Model4GetFields_test_get_fields_from_model(models.Model):
            integer = models.IntegerField()
        fields = get_fields_from_model(Model4GetFields_test_get_fields_from_model)
        assert get_field_by_name_or_raise(Model4GetFields_test_get_fields_from_model, 'id') in fields
        assert get_field_by_name_or_raise(Model4GetFields_test_get_fields_from_model, 'integer') in fields

    def test_get_local_fields(self):
        class ModelForGetLocalFields_test_get_local_fields(models.Model):
            integer = models.IntegerField()
        fields = get_local_fields(ModelForGetLocalFields_test_get_local_fields)
        assert get_field_by_name_or_raise(ModelForGetLocalFields_test_get_local_fields, 'id') in fields
        assert get_field_by_name_or_raise(ModelForGetLocalFields_test_get_local_fields, 'integer') in fields

    def test_get_field_names_of_model(self):
        class Model4GetFieldNames_test_get_field_names_of_model(models.Model):
            smallinteger = models.SmallIntegerField()
        fields = get_field_names_of_model(Model4GetFieldNames_test_get_field_names_of_model)
        assert 'smallinteger' in fields
        assert 'unknown' not in fields

    def test_get_many_to_many_fields_from_model(self):
        class ModelRelated_test_get_many_to_many_fields_from_model(models.Model): pass
        class ModelWithM2M_test_get_many_to_many_fields_from_model(models.Model):
            manytomany = models.ManyToManyField('ModelRelated_test_get_many_to_many_fields_from_model', related_name='m2m')
        fields = get_many_to_many_fields_from_model(ModelWithM2M_test_get_many_to_many_fields_from_model)
        assert get_field_by_name_or_raise(ModelWithM2M_test_get_many_to_many_fields_from_model, 'manytomany') in fields
        assert get_field_by_name_or_raise(ModelWithM2M_test_get_many_to_many_fields_from_model, 'id') not in fields

    def test_is_model_class(self):
        class MyModel_test_is_model_class(models.Model): pass
        assert is_model_class(MyModel_test_is_model_class) == True
        class X(object): pass
        assert is_model_class(X) == False

    def test_is_model_abstract(self):
        class AbstractModel_test_is_model_abstract(models.Model):
            class Meta:
                abstract = True
        assert is_model_abstract(AbstractModel_test_is_model_abstract)

        class ConcreteModel_test_is_model_abstract(models.Model):
            class Meta:
                abstract = False
        assert is_model_abstract(ConcreteModel_test_is_model_abstract) == False

    def test_is_model_managed(self):
        class NotManagedModel_test_is_model_managed(models.Model):
            class Meta:
                managed = False
        assert is_model_managed(NotManagedModel_test_is_model_managed) == False

        class ManagedModel_test_is_model_managed(models.Model):
            class Meta:
                managed = True
        assert is_model_managed(ManagedModel_test_is_model_managed)

    def test_model_has_the_field(self):
        class ModelWithWithoutFields_test_model_has_the_field(models.Model):
            integer = models.IntegerField()
            selfforeignkey = models.ForeignKey('self', null=True, on_delete=models.DO_NOTHING)
            manytomany = models.ManyToManyField('self', related_name='m2m')
        assert model_has_the_field(ModelWithWithoutFields_test_model_has_the_field, 'integer')
        assert model_has_the_field(ModelWithWithoutFields_test_model_has_the_field, 'selfforeignkey')
        assert model_has_the_field(ModelWithWithoutFields_test_model_has_the_field, 'manytomany')
        assert model_has_the_field(ModelWithWithoutFields_test_model_has_the_field, 'x') == False


class DjangoHelperFieldsTest(TestCase):
    def test_get_unique_field_name(self):
        class Model4GetUniqueFieldName_test_get_unique_field_name(models.Model):
            integer = models.IntegerField()
        field = get_field_by_name_or_raise(Model4GetUniqueFieldName_test_get_unique_field_name, 'integer')
        assert get_unique_field_name(field) == 'django_dynamic_fixture.tests.test_django_helper.Model4GetUniqueFieldName_test_get_unique_field_name.integer'

    def test_get_related_model(self):
        class ModelRelated_test_get_related_model(models.Model): pass
        class Model4GetRelatedModel_test_get_related_model(models.Model):
            fk = models.ForeignKey(ModelRelated_test_get_related_model, on_delete=models.DO_NOTHING)
        assert get_related_model(get_field_by_name_or_raise(Model4GetRelatedModel_test_get_related_model, 'fk')) == \
               ModelRelated_test_get_related_model


    def test_field_is_a_parent_link(self):
        class ModelParent_test_get_related_model(models.Model): pass
        class Model4FieldIsParentLink_test_get_related_model(ModelParent):
            o2o_with_parent_link = models.OneToOneField(ModelParent_test_get_related_model, parent_link=True, related_name='my_custom_ref_x', on_delete=models.DO_NOTHING)
        class Model4FieldIsParentLink2(ModelParent):
            o2o_without_parent_link = models.OneToOneField(ModelParent_test_get_related_model, parent_link=False, related_name='my_custom_ref_y', on_delete=models.DO_NOTHING)
        # FIXME
        # assert field_is_a_parent_link(get_field_by_name_or_raise(Model4FieldIsParentLink, 'o2o_with_parent_link'))
        assert field_is_a_parent_link(get_field_by_name_or_raise(Model4FieldIsParentLink2, 'o2o_without_parent_link')) == False

    def test_field_has_choices(self):
        class Model4FieldHasChoices_test_get_related_model(models.Model):
            with_choices = models.IntegerField(choices=((1, 1), (2, 2)))
            without_choices = models.IntegerField()
        assert field_has_choices(get_field_by_name_or_raise(Model4FieldHasChoices_test_get_related_model, 'with_choices'))
        assert field_has_choices(get_field_by_name_or_raise(Model4FieldHasChoices_test_get_related_model, 'without_choices')) == False

    def test_field_has_default_value(self):
        class Model4FieldHasDefault_test_field_has_default_value(models.Model):
            with_default = models.IntegerField(default=1)
            without_default = models.IntegerField()
        assert field_has_default_value(get_field_by_name_or_raise(Model4FieldHasDefault_test_field_has_default_value, 'with_default'))
        assert field_has_default_value(get_field_by_name_or_raise(Model4FieldHasDefault_test_field_has_default_value, 'without_default')) == False

    def test_field_is_unique(self):
        class Model4FieldMustBeUnique_test_field_is_unique(models.Model):
            unique = models.IntegerField(unique=True)
            not_unique = models.IntegerField()
        assert field_is_unique(get_field_by_name_or_raise(Model4FieldMustBeUnique_test_field_is_unique, 'unique'))
        assert field_is_unique(get_field_by_name_or_raise(Model4FieldMustBeUnique_test_field_is_unique, 'not_unique')) == False

    def test_is_key_field(self):
        class ModelForKeyField_test_is_key_field(models.Model):
            integer = models.IntegerField()
        assert is_key_field(get_field_by_name_or_raise(ModelForKeyField_test_is_key_field, 'id'))
        assert is_key_field(get_field_by_name_or_raise(ModelForKeyField_test_is_key_field, 'integer')) == False

    def test_is_relationship_field(self):
        class ModelForRelationshipField_test_is_relationship_field(models.Model):
            fk = models.ForeignKey('self', on_delete=models.DO_NOTHING)
            one2one = models.OneToOneField('self', on_delete=models.DO_NOTHING)
        assert is_relationship_field(get_field_by_name_or_raise(ModelForRelationshipField_test_is_relationship_field, 'fk'))
        assert is_relationship_field(get_field_by_name_or_raise(ModelForRelationshipField_test_is_relationship_field, 'one2one'))
        assert is_relationship_field(get_field_by_name_or_raise(ModelForRelationshipField_test_is_relationship_field, 'id')) == False

    def test_is_file_field(self):
        class ModelForFileField_test_is_file_field(models.Model):
            filefield = models.FileField()
        assert is_file_field(get_field_by_name_or_raise(ModelForFileField_test_is_file_field, 'filefield'))
        assert is_file_field(get_field_by_name_or_raise(ModelForFileField_test_is_file_field, 'id')) == False


class PrintFieldValuesTest(TestCase):
    def test_model_not_saved_do_not_raise_an_exception(self):
        instance = N(ModelWithNumbers)
        print_field_values(instance)

    def test_model_saved_do_not_raise_an_exception(self):
        instance = G(ModelWithNumbers)
        print_field_values(instance)

    def test_print_accept_list_of_models_too(self):
        instances = G(ModelWithNumbers, n=2)
        print_field_values(instances)
        print_field_values([G(ModelWithNumbers), G(ModelWithNumbers)])

    def test_print_accept_a_queryset_too(self):
        G(ModelWithNumbers, n=2)
        print_field_values(ModelWithNumbers.objects.all())




from pathlib import Path
from setuptools import setup, find_packages

VERSION = '4.0.1'

tests_require = []

install_requires = []

this_directory = Path(__file__).parent
long_description = (this_directory / 'README.md').read_text()

setup(name='django-dynamic-fixture',
      url='https://github.com/paulocheque/django-dynamic-fixture',
      author="paulocheque",
      author_email='paulocheque@gmail.com',
      keywords='python django testing fixture',
      description='A full library to create dynamic model instances for testing purposes.',
      long_description_content_type='text/markdown',
      long_description=long_description,
      license='MIT',
      classifiers=[
          'Framework :: Django',
          'Operating System :: OS Independent',
          'Topic :: Software Development',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Programming Language :: Python :: Implementation :: PyPy',
      ],

      version=VERSION,
      install_requires=install_requires,
      tests_require=tests_require,
      test_suite='pytest',
      extras_require={'test': tests_require},

      packages=find_packages(),
)




import re

from django.test import TestCase
import pytest

from django_dynamic_fixture.models_test import *
from django_dynamic_fixture.ddf import *
from django_dynamic_fixture.fixture_algorithms.sequential_fixture import SequentialDataFixture


data_fixture = SequentialDataFixture()


class DDFTestCase(TestCase):
    def setUp(self):
        self.ddf = DynamicFixture(data_fixture)
        DDFLibrary.get_instance().clear()


class TeachAndLessonsTest(DDFTestCase):
    def test_teach_a_default_lesson_for_a_model(self):
        self.ddf.teach(ModelForLibrary, integer=1000)
        instance = self.ddf.get(ModelForLibrary)
        assert instance.integer == 1000

    def test_default_lesson_may_be_overrided_although_it_is_an_anti_pattern(self):
        self.ddf.teach(ModelForLibrary, integer=1000)
        instance = self.ddf.get(ModelForLibrary)
        assert instance.integer == 1000
        self.ddf.teach(ModelForLibrary, integer=1001)
        instance = self.ddf.get(ModelForLibrary)
        assert instance.integer == 1001

    def test_it_must_NOT_raise_an_error_if_user_try_to_use_a_not_saved_default_configuration(self):
        self.ddf.get(ModelForLibrary)

    def test_it_must_raise_an_error_if_try_to_set_a_static_value_to_a_field_with_unicity(self):
        with pytest.raises(InvalidConfigurationError):
            self.ddf.teach(ModelForLibrary, integer_unique=1000)

    def test_it_allows_to_use_masks_as_lessons_for_unique_integer_fields(self):
        self.ddf.teach(ModelForLibrary, integer_unique=Mask('1###'))
        instance = self.ddf.get(ModelForLibrary)
        assert 1000 <= int(instance.integer_unique) <= 1999

    def test_it_allows_to_use_masks_as_lessons_for_unique_char_fields(self):
        self.ddf.teach(ModelWithUniqueCharField, text_unique=Mask('---- ### __'))
        instance = self.ddf.get(ModelWithUniqueCharField)
        assert re.match(r'[A-Z]{4} [0-9]{3} [a-z]{2}', instance.text_unique)

    def test_it_must_accept_dynamic_values_for_fields_with_unicity(self):
        self.ddf.teach(ModelForLibrary, integer_unique=lambda field: 1000)

    def test_it_must_NOT_propagate_lessons_for_internal_dependencies(self):
        self.ddf.teach(ModelForLibrary, foreignkey=DynamicFixture(data_fixture, integer=1000))
        instance = self.ddf.get(ModelForLibrary)
        assert instance.integer != 1000
        assert instance.foreignkey.integer == 1000

    def test_it_must_use_lessons_for_internal_dependencies(self):
        # ModelForLibrary.foreignkey is a `ModelForLibrary2`
        self.ddf.teach(ModelForLibrary, integer=1000)
        self.ddf.teach(ModelForLibrary2, integer=1001)
        instance = self.ddf.get(ModelForLibrary, foreignkey=DynamicFixture(data_fixture))
        assert instance.integer == 1000
        assert instance.foreignkey.integer == 1001

    def test_it_uses_lessons_for_base_model_when_creating_a_proxy_model(self):
        self.ddf.teach(ModelForLibrary, integer=123)
        instance = self.ddf.get(ProxyModelForLibrary)
        assert instance.__class__ is ProxyModelForLibrary
        assert instance.integer == 123

    def test_it_uses_lessons_for_proxy_models_when_creating_the_base_model(self):
        self.ddf.teach(ProxyModelForLibrary, integer=456)
        instance = self.ddf.get(ModelForLibrary)
        assert instance.__class__ is ModelForLibrary
        assert instance.integer == 456

    def test_it_uses_lessons_for_proxy_models_when_creating_the_proxy_model(self):
        self.ddf.teach(ProxyModelForLibrary, integer=789)
        instance = self.ddf.get(ProxyModelForLibrary)
        assert instance.__class__ is ProxyModelForLibrary
        assert instance.integer == 789

    # Not implemented yet
    # def test_teaching_must_store_ddf_configs_too(self):
    #     self.ddf.teach(ModelForLibrary, fill_nullable_fields=False)
    #     instance = self.ddf.get(ModelForLibrary)
    #     assert instance.integer is None

    #     DDFLibrary.get_instance().clear()
    #     self.ddf.teach(ModelForLibrary, fill_nullable_fields=True)
    #     instance = self.ddf.get(ModelForLibrary)
    #     assert instance.integer is not None

    # Not implemented yet
    # def test_teaching_ddf_configs_must_NOT_be_propagated_to_another_models(self):
    #     self.ddf.teach(ModelForLibrary, fill_nullable_fields=False)
    #     instance = self.ddf.get(ModelForLibrary)
    #     assert instance.integer is None
    #     assert instance.foreignkey.integer is None

    #     DDFLibrary.get_instance().clear()
    #     self.ddf.teach(ModelForLibrary, fill_nullable_fields=True)
    #     instance = self.ddf.get(ModelForLibrary)
    #     assert instance.integer is not None
    #     assert instance.foreignkey.integer is None # not populated


class TeachingAndCustomLessonsTest(DDFTestCase):
    def test_a_model_can_have_custom_lessons(self):
        self.ddf.teach(ModelForLibrary, integer=1000, ddf_lesson=None)
        self.ddf.teach(ModelForLibrary, integer=1001, ddf_lesson='a name')
        instance = self.ddf.get(ModelForLibrary)
        assert instance.integer == 1000
        instance = self.ddf.get(ModelForLibrary, ddf_lesson='a name')
        assert instance.integer == 1001

    def test_custom_lessons_must_not_be_used_if_not_explicity_specified(self):
        self.ddf.teach(ModelForLibrary, integer=1000, ddf_lesson='a name')
        instance = self.ddf.get(ModelForLibrary)
        assert instance.integer != 1000

    def test_a_model_can_have_many_custom_lessons(self):
        self.ddf.teach(ModelForLibrary, integer=1000, ddf_lesson='a name')
        self.ddf.teach(ModelForLibrary, integer=1001, ddf_lesson='a name 2')

        instance = self.ddf.get(ModelForLibrary, ddf_lesson='a name')
        assert instance.integer == 1000

        instance = self.ddf.get(ModelForLibrary, ddf_lesson='a name 2')
        assert instance.integer == 1001

    def test_it_must_raise_an_error_if_user_try_to_use_a_not_saved_configuration(self):
        with pytest.raises(InvalidConfigurationError):
            self.ddf.get(ModelForLibrary, ddf_lesson='a not teached lesson')

    def test_default_lesson_and_custom_lesson_must_work_together(self):
        # regression test
        self.ddf.teach(ModelForLibrary, integer=1000, ddf_lesson='a name')
        self.ddf.teach(ModelForLibrary, integer=1001, ddf_lesson=True)
        self.ddf.teach(ModelForLibrary, integer=1002, ddf_lesson='a name2')
        instance = self.ddf.get(ModelForLibrary, ddf_lesson='a name')
        assert instance.integer == 1000
        instance = self.ddf.get(ModelForLibrary)
        assert instance.integer == 1001
        instance = self.ddf.get(ModelForLibrary, ddf_lesson='a name2')
        assert instance.integer == 1002

    def test_default_lesson_and_custom_lesson_must_work_together_for_different_models(self):
        # regression test
        self.ddf.teach(ModelForLibrary, integer=1000, ddf_lesson='a name')
        self.ddf.teach(ModelForLibrary, integer=1001, ddf_lesson=True)
        self.ddf.teach(ModelForLibrary, integer=1002, ddf_lesson='a name2')
        self.ddf.teach(ModelForLibrary2, integer=2000, ddf_lesson='a name')
        self.ddf.teach(ModelForLibrary2, integer=2001, ddf_lesson=True)
        self.ddf.teach(ModelForLibrary2, integer=2002, ddf_lesson='a name2')

        instance = self.ddf.get(ModelForLibrary, ddf_lesson='a name')
        assert instance.integer == 1000
        instance = self.ddf.get(ModelForLibrary)
        assert instance.integer == 1001
        instance = self.ddf.get(ModelForLibrary, ddf_lesson='a name2')
        assert instance.integer == 1002

        instance = self.ddf.get(ModelForLibrary2, ddf_lesson='a name')
        assert instance.integer == 2000
        instance = self.ddf.get(ModelForLibrary2)
        assert instance.integer == 2001
        instance = self.ddf.get(ModelForLibrary2, ddf_lesson='a name2')
        assert instance.integer == 2002


class DDFLibraryTest(TestCase):
    def setUp(self):
        self.lib = DDFLibrary()

    def test_add_and_get_configuration_without_string_name(self):
        self.lib.add_configuration(ModelForLibrary, {'a': 1})
        assert self.lib.get_configuration(ModelForLibrary) == {'a': 1}
        assert self.lib.get_configuration(ModelForLibrary, name=DDFLibrary.DEFAULT_KEY) == {'a': 1}
        assert self.lib.get_configuration(ModelForLibrary, name=None) == {'a': 1}

        self.lib.clear()
        self.lib.add_configuration(ModelForLibrary, {'a': 2}, name=None)
        assert self.lib.get_configuration(ModelForLibrary) == {'a': 2}
        assert self.lib.get_configuration(ModelForLibrary, name=DDFLibrary.DEFAULT_KEY) == {'a': 2}
        assert self.lib.get_configuration(ModelForLibrary, name=None) == {'a': 2}

        self.lib.clear()
        self.lib.add_configuration(ModelForLibrary, {'a': 3}, name=True)
        assert self.lib.get_configuration(ModelForLibrary) == {'a': 3}
        assert self.lib.get_configuration(ModelForLibrary, name=DDFLibrary.DEFAULT_KEY) == {'a': 3}
        assert self.lib.get_configuration(ModelForLibrary, name=None) == {'a': 3}

    def test_add_and_get_configuration_with_name(self):
        self.lib.add_configuration(ModelForLibrary, {'a': 1}, name='x')
        assert self.lib.get_configuration(ModelForLibrary, name='x') == {'a': 1}

    def test_clear_config(self):
        self.lib.clear_configuration(ModelForLibrary) # run ok if empty
        self.lib.add_configuration(ModelForLibrary, {'a': 1})
        self.lib.add_configuration(ModelForLibrary, {'a': 2}, name='x')
        self.lib.add_configuration(ModelForLibrary2, {'a': 3})
        self.lib.clear_configuration(ModelForLibrary)
        assert self.lib.get_configuration(ModelForLibrary) == {}
        with pytest.raises(Exception):
            self.lib.get_configuration(ModelForLibrary, name='x')
        assert self.lib.get_configuration(ModelForLibrary2) == {'a': 3}

    def test_clear(self):
        self.lib.add_configuration(ModelForLibrary, {'a': 1})
        self.lib.add_configuration(ModelForLibrary, {'a': 2}, name='x')
        self.lib.add_configuration(ModelForLibrary2, {'a': 3})
        self.lib.add_configuration(ModelForLibrary2, {'a': 4}, name='x')
        self.lib.clear()
        assert self.lib.get_configuration(ModelForLibrary) == {}
        with pytest.raises(Exception):
            self.lib.get_configuration(ModelForLibrary, name='x')
        assert self.lib.get_configuration(ModelForLibrary2) == {}
        with pytest.raises(Exception):
            self.lib.get_configuration(ModelForLibrary2, name='x')
